#!/usr/bin/env python
# -*- coding: utf-8  -*-

# USO CORRECTO: sudo python script.py [-v] [-h|-c|-p <num>]
#   * -v -> modo chivato activado
#   * -h o --help -> ayuda
#   * -c o --clean -> limpieza previa
#   * -p <n> o --partial <n> -> comprobación intermedia
#                               número n
#   * -s o --show -> muestra los ejercicios a realizar
# Debe ser ejecutado con sudo o bajo permisos de admi-
# nistrador o no funcionará correctamente

# Creado por Noelia Sales Montes

# This program is free software: you can redistribute
# it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software
# Foundation, either version 3 of the License, or any
# later version.

# This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

# You should have received a copy of the GNU General 
# Public License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

# Contacto:  
#            noelia.salesmontes@gmail.com 

"""
Script de comprobación del proyecto WikiUNIX
Comprueba la resolución de ejercicios
"""

import os
import getopt
import sys
import filecmp
import stat
import re
import shutil

# Directorios de comprobación
__base__ = ".correccion/base"
__clean__ = ".correccion/limpia"
__partial__ = ".correccion/parcial"
__final__ = ".correccion/final"
# Fichero que contiene datos a mostrar
__exercises__ = ".correccion/ejercicios.dat"
__usage__ = ".correccion/uso.dat"

# Tipos de ficheros de comprobación
__direxists__ = "dir_existentes.dat"
__fileexists__ = "fich_existentes.dat"
__dirclean__ = "dir_eliminados.dat"
__fileclean__ = "fich_eliminados.dat"
__perm__ = "permisos.dat"
__own__ = "propietarios.dat"
__grp__ = "grupos.dat"
__mask__ = "mascara.dat"
__fileequals__ = "fich_iguales.dat"
__direquals__ = "dir_iguales.dat"
__content__ = "contenido.dat"
__links__ = "enlaces.dat"

# Tipos de ficheros de acciones
__dirremove__ = "dir_a_borrar.dat"
__fileremove__ = "fich_a_borrar.dat"
__changeperm__ = "cambiar_permisos.dat"
__changeown__ = "cambiar_propietarios.dat"
__changegrp__ = "cambiar_grupos.dat"
__changemask__ = "cambiar_mascara.dat"
__filerestore__ = "fich_a_restaurar.dat"
__dirrestore__ = "dir_a_restaurar.dat"
__commands__ = "ordenes.dat"

# Funciones y modos
__verbose__ = False
__fun__ = __final__

# ----------------------------------------------------------------------
#                        FUNCIONES AUXILIARES
# ----------------------------------------------------------------------

def usage():
    """
    Impresión de la utilización del script
    """
    if os.path.exists(__usage__):
        with open(__usage__, 'r') as __f:
            for __l in __f.read().split('\n'):
                print __l
        __f.close()
    else:
        error("No existe el fichero de muestra de uso")

def error(problem):
    """
    Función auxiliar para simplificar la salir con error
    """
    print "Error: %s" % problem
    sys.exit(2)

def verb(__line):
    """
    Función auxiliar para simplificar el modo chivato
    """
    if __verbose__ == True:
        print __line

def show():
    """
    Función auxiliar que se encarga de mostrar todos los ejercicios relacionados
    con este script
    """
    if os.path.exists(__exercises__):
        with open(__exercises__, 'r') as __f:
            for __l in __f.read().split('\n'):
                print __l
        __f.close()
    else:
        error("No existe el fichero de muestra de ejercicios")

def remove(__file):
    """
    Función auxiliar para simplificar el borrado de un único fichero
    """
    if os.path.exists(__file):
        try:
            os.remove(__file)
            verb("   -> Borrando fichero %s <-" % __file)
        except os.error:
            verb("No se ha podido borrar el fichero %s" % __file)

def to_decimal(__int):
    """
    Conversión de una cadena en octal a un número en decimal
    """
    ind = len(__int) - 1
    dec = 0
    for __l in __int:
        dec += int(__l) * 8**ind
        ind -= 1
    return dec

def removedir(__dir):
    """
    Función auxiliar para simplificar el borrado de un directorio
    """
    if os.path.exists(__dir):
        try:
            os.removedirs(__dir)
            verb("    -> Borrando directorio %s <-" % __dir)
        except os.error:
            verb("No se ha podido borrar el directorio %s" % __dir)


def direxists(__dir):
    """
    Función auxiliar para simplificar la comprobación de existencia de un
    directorio
    """
    if os.path.exists(__dir) and os.path.isdir(__dir):
        return True
    else:
        return False
    
def notdirexists(__dir):
    """
    Función auxiliar para simplificar la comprobación de no existencia de un
    directorio
    """
    return not direxists(__dir)

def fileexists(__file):
    """
    Función auxiliar para simplificar la comprobación de existencia de un
    fichero
    """
    if os.path.exists(__file) and os.path.isfile(__file):
        return True
    else:
        return False
    
def notfileexists(__file):
    """
    Función auxiliar para simplificar la comprobación de no existencia 
    de un fichero
    """
    return not fileexists(__file)


def comp_perm(__elem, __perm):
    """
    Comprueba si un elemento (fichero o directorio) tiene unos determinados
    permisos (estrictamente)
    """
    if os.path.exists(__elem):
        try:
            # Permisos en octal y en forma de cadena
            mode = str(oct(os.stat(__elem)[stat.ST_MODE]))
        except os.error:
            error("No se han podido leer los permisos de %s" % __elem)
            
        if mode == __perm:
            return True
        else:
            return False
    else:
        verb("No existe %s" % __elem)
        return False

def comp_user(__elem, __user):
    """
    Comprueba si un elemento (fichero o directorio) es poseído por un usuario
    en función de su identificador
    """
    if os.path.exists(__elem):
        try:
            __st = os.stat(__elem)
            __uid = __st[stat.ST_UID]
        except os.error:
            error("No se ha podido leer el usuario de %s" % __elem)

        if __uid == __user:
            return True
        else:
            return False
    else:
        return False

def comp_grp(__elem, __grp):
    """
    Comprueba si un elemento (fichero o directorio) es poseído por un grupo en
    función de su identificador
    """
    if os.path.exists(__elem):
        try:
            __st = os.stat(__elem)
            gid = __st[stat.ST_GID]
        except os.error:
            error("No se ha podido leer el grupo de %s" % __elem)

        if gid == __grp:
            return True
        else:
            return False
    else:
        return False

def comp_mask(__mask):
    """
    Comprueba la máscara de permisos del sistema
    """
    try:
        mas = os.umask(__mask)
    except os.error:
        error("No se ha podido leer la máscara de permisos")
    
    if mas == __mask:
        return True
    else:
        os.umask(__mask)
        return False

def comp_content(__fich, __er):
    """
    Comprueba que una determinada expresión regular se encuentra en una de las
    líneas del fichero
    """
    if os.path.exists(__fich):
        with open(__fich, 'r') as __f:
            for __l in __f.read().split('\n'):
                exp = re.compile(__er)
                match = exp.match(__l)
                if match:
                    return True       
        __f.close()
        # Si no se encuentra coincidencia
        return False

def comp_link(__elem, __src):
    """
    Comprueba el enlace de un fichero
    """    
    if os.path.exists(__elem):
        return (os.readlink(__elem) == __src)
        

def chg_own(__elem, __user):
    """
    Cambia el propietario de un elemento (fichero o directorio)
    """
    if os.path.exists(__elem):
        __st = os.stat(__elem)
        gid = __st[stat.ST_GID]
        try:
            os.chown(__elem, int(__user), gid)
        except os.error:
            error("No se ha podido modificar el usuario de %s" % __elem)
            
def chg_grp(__elem, __grp):
    """
    Cambia el grupo de un elemento (fichero o directorio)
    """
    if os.path.exists(__elem):
        __st = os.stat(__elem)
        uid = __st[stat.ST_UID]
        try:
            os.chown(__elem, uid, int(__grp))
        except os.error:
            error("No se ha podido modificar el usuario de %s" % __elem)

def chg_perm(__elem, __perm):
    """
    Cambiar los permisos de un elemento (fichero o directorio)
    """
    if os.path.exists(__elem):
        try:
            os.chmod(__elem, to_decimal(__perm))
        except os.error:
            error("No se han podido modificar los permisos de %s" % __elem)

def chg_mask(__mask):
    """
    Cambia la máscara de permisos de archivos
    """
    try:
        os.umask(int(__mask))
    except os.error:
        error("No se ha podido modificar la máscara del sistema")


def check_simple(__funcion, __actual, __mensaje, __verbose):
    """
    Comprobación básica que recibe una función y un único argumento
    """
    if fileexists(__actual):
        verb("%s (%s)" % (__verbose, __actual))
        # Abrimos el fichero para realizar la comprobación pertinente
        with open(__actual, 'r') as __f:
            for __l in __f.read().split('\n'):
                if not __funcion(__l):
                    error("%s %s" % (__mensaje, __l))                
        __f.close()

def check_doble(__funcion, __actual, __mensaje, __verbose):
    """
    Comprobación doble que recibe una función y dos argumentos
    """
    if fileexists(__actual):
        verb("%s (%s)" % (__verbose, __actual))
        # Abrimos el fichero para realizar la comprobación pertinente
        with open(__actual, 'r') as __f:
            for __l in __f.read().split('\n'):
                __elema = __l[:__l.find(' ')]
                __elemb = __l[__l.find(' ') + 1:]
                if not __funcion(__elema, __elemb):
                    error("%s %s (%s)" % (__mensaje, __elema, __elemb))
        __f.close()

        
def change_simple(__funcion, __actual, __mensaje, __verbose):
    """
    Modificación simple que recibe una función y un único argumento
    """
    if fileexists(__actual):
        verb("%s (%s)" % (__verbose, __actual))
        with open(__actual, 'r') as __f:
            for __l in __f.read().split('\n'):
                __funcion(__l)
        __f.close()

def change_doble(__funcion, __actual, __mensaje, __verbose):
    """
    Modificación doble que recibe una función y dos argumentos
    """
    if fileexists(__actual):
        verb("%s (%s)" % (__verbose, __actual))
        with open(__actual, 'r') as __f:
            for __l in __f.read().split('\n'):
                __elema = __l[:__l.find(' ')]
                __elemb = __l[__l.find(' ') + 1:]
                __funcion(__elema, __elemb)
        __f.close()
            

# ----------------------------------------------------------------------
#              FUNCIÓN GENERAL DE LIMPIEZA Y COMPROBACIÓN
# ----------------------------------------------------------------------
def check(__comp):
    """
    Se pasa el nombre del tipo de comprobación a realizar (que debe
    corresponderse al directorio de comprobación adecuado) y se ejecuta dicha
    verificación
    """
    # Comprobaciones de existencia
    check_simple(direxists, os.path.join(__comp, __direxists__),
                 "No existe el directorio",
                 "  - Comprobación de existencia de directorios")
    
    check_simple(fileexists, os.path.join(__comp, __fileexists__),
                 "No existe el fichero",
                 "  - Comprobación de existencia de ficheros")

    check_simple(notdirexists, os.path.join(__comp, __dirclean__),
                 "Aún existe el directorio",
                 "  - Comprobación de no existencia de directorios")
    
    check_simple(notfileexists, os.path.join(__comp, __fileclean__),
                 "Aún existe el fichero",
                 "  - Comprobación de no existencia de ficheros")

    # Comprobaciones de permisos
    check_doble(comp_perm, os.path.join(__comp, __perm__),
                "No coinciden los permisos de",
                "  - Comprobación de los permisos de los ficheros")

    check_doble(comp_user, os.path.join(__comp, __own__),
                "No coinciden los permisos de",
                "  - Comprobación del propietario de los ficheros")

    check_doble(comp_grp, os.path.join(__comp, __grp__),
                "No coinciden los permisos de",
                "  - Comprobación del grupo de los ficheros")

    check_simple(comp_mask, os.path.join(__comp, __mask__),
                 "No coincide la máscara del sistema -",
                 "  - Comprobación de la máscara del sistema")

    check_doble(comp_content, os.path.join(__comp, __content__),
                "No se encuentra coincidencia",
                "  - Comprobación de contenido de ficheros")
    
    # Comprobación de enlaces
    check_doble(comp_link, os.path.join(__comp, __links__),
                "No coincide el enlace de ",
                "  - Comprobación de enlaces")
    
    # Comprobaciones de igualdades
    check_doble(filecmp.cmp, os.path.join(__comp, __fileequals__),
                "No coincide el fichero",
                 "  - Comprobación de ficheros adicionales")
    
    check_doble(filecmp.dircmp, os.path.join(__comp, __direquals__),
                "No coindice el directorio",
                "  - Comprobación de directorios adicionales")

    # Borrados
    change_simple(removedir, os.path.join(__comp, __dirremove__),
                  "Borrando directorio",
                  "  - Borrado de directorios")
    
    change_simple(remove, os.path.join(__comp, __fileremove__),
                  "Borrando fichero",
                  "  - Borrado de ficheros")

    # Restaurados
    change_doble(shutil.copy, os.path.join(__comp, __filerestore__),
                  "Restaurando fichero",
                  "  - Restaurando ficheros")
    
    change_doble(shutil.copytree, os.path.join(__comp, __dirrestore__),
                  "Restaurando directorio",
                  "  - Restaurando directorios")

    # Modificaciones de permisos
    change_doble(chg_perm, os.path.join(__comp, __changeperm__),
                 "Modificando permisos",
                 "  - Modificación de permisos")
    
    change_doble(chg_own, os.path.join(__comp, __changeown__),
                 "Modificando propietario de ",
                 "  - Modificación de propietarios")
    
    change_doble(chg_grp, os.path.join(__comp, __changegrp__),
                 "Modificando grupo propietario de ",
                 "  - Modificación de grupos")
    
    change_simple(chg_mask, os.path.join(__comp, __changemask__),
                  "Modificando máscara",
                  "  - Modificación de la máscara del sistema")

    change_simple(os.popen, os.path.join(__comp, __commands__),
                  "Ejecutando orden",
                  "  - Ejecución de órdenes")


# ----------------------------------------------------------------------
#                           FUNCIÓN PRINCIPAL
# ----------------------------------------------------------------------
if __name__ == "__main__":
    try:
        __opts__, __args__ = getopt.getopt(sys.argv[1:],
                                       "vhcp:s",
                                       ["__verbose__","help", "clean",
                                        "partial","show"])
    except getopt.GetoptError, err:
        # Imprimir error y ayuda y salir
        print str(err)
        usage()
        sys.exit(2)

    # Identificar opciones
    for o, a in __opts__:
        if o == "-v":
            __verbose__ = True
        elif o in ("-h", "--help"):
            usage()
            sys.exit()
        elif o in ("-c", "--clean"):  
            __fun__ = __clean__
        elif o in ("-p", "--partial"):
            __fun__ = __partial__ + a
            print "PARCIAL: %s" % __fun__
        elif o in ("-s", "--show"):
            show()
            sys.exit()
        else:
            assert False, "Opción no manejada"

    # Comprobaciones
    check(__base__)
    check(__fun__)

    print "Verificación correcta"
    # Mensajes finales
    if __fun__ == __clean__:
        print "  -> Puedes iniciar el desarrollo de los ejercicios"
    if __fun__ == __final__:
        print "  -> Has completado los ejercicios: ¡BIEN HECHO!"
